#!/usr/bin/env python3

# Internal packages
import sys
import unittest
from pathlib import Path
from typing import List, Tuple, Final

# Append the module path for md2anki
sys.path.append(str(Path(__file__).parent.parent.joinpath("src")))

# Local modules
from md2anki.cli import (
    parse_cli_args,
    Md2AnkiArgs,
    MdInputFileNotFoundException,
)


class TestCliArgs(unittest.TestCase):
    def setUp(self):
        self.valid_args = [__file__]

        self.cli_args: List[List[str]] = list()
        self.results: List[Md2AnkiArgs] = list()
        self.expected: List[Md2AnkiArgs] = list()

        test_data: List[Tuple[List[str], Md2AnkiArgs]] = [
            (
                [__file__],
                Md2AnkiArgs(
                    md_input_file_paths=[Path(__file__)],
                    md_output_file_paths=[Path(__file__)],
                    anki_output_file_path=Path(__file__).parent.joinpath(
                        f"{Path(__file__).stem}.apkg"
                    ),
                    custom_program=parse_cli_args([__file__]).custom_program,
                    custom_program_args=parse_cli_args([__file__]).custom_program_args,
                ),
            ),
            (
                [__file__, "-file-dir", str(Path(__file__).parent)],
                Md2AnkiArgs(
                    md_input_file_paths=[Path(__file__)],
                    md_output_file_paths=[Path(__file__)],
                    additional_file_dirs=[Path(__file__).parent],
                    anki_output_file_path=Path(__file__).parent.joinpath(
                        f"{Path(__file__).stem}.apkg"
                    ),
                    custom_program=parse_cli_args([__file__]).custom_program,
                    custom_program_args=parse_cli_args([__file__]).custom_program_args,
                ),
            ),
            (
                [__file__, str(Path(__file__).parent.joinpath("__init__.py"))],
                Md2AnkiArgs(
                    md_input_file_paths=[
                        Path(__file__),
                        Path(__file__).parent.joinpath("__init__.py"),
                    ],
                    md_output_file_paths=[
                        Path(__file__),
                        Path(__file__).parent.joinpath("__init__.py"),
                    ],
                    anki_output_file_path=Path(__file__).parent.joinpath(
                        f"{Path(__file__).stem}.apkg"
                    ),
                    custom_program=parse_cli_args([__file__]).custom_program,
                    custom_program_args=parse_cli_args([__file__]).custom_program_args,
                ),
            ),
            (
                [
                    __file__,
                    "-custom-program",
                    "py",
                    "python3.8",
                    "-custom-program",
                    "cpp",
                    "gcc",
                ],
                Md2AnkiArgs(
                    md_input_file_paths=[Path(__file__)],
                    md_output_file_paths=[Path(__file__)],
                    anki_output_file_path=Path(__file__).parent.joinpath(
                        f"{Path(__file__).stem}.apkg"
                    ),
                    custom_program={
                        **parse_cli_args([__file__]).custom_program,
                        **{"py": ["python3.8"], "cpp": ["gcc"]},
                    },
                    custom_program_args=parse_cli_args([__file__]).custom_program_args,
                ),
            ),
        ]

        for test_cli_args, test_expected in test_data:
            self.cli_args.append(test_cli_args)
            self.results.append(parse_cli_args(test_cli_args))
            self.expected.append(test_expected)

    def test_cli_logic(self):
        # Detect not existing files
        file_path_not_found: Final = Path("not_found.md")
        error_file_not_found: Final = parse_cli_args([str(file_path_not_found)]).error
        if error_file_not_found is not None and isinstance(
            error_file_not_found, MdInputFileNotFoundException
        ):
            self.assertEqual(
                error_file_not_found.md_input_file_path,
                file_path_not_found,
                "Error is thrown if file can not be found",
            )
        else:
            self.fail(f"Expected file not found error ({error_file_not_found=})")
        # Detect not existing additional file directories
        file_dir_not_found: Final = Path("not_found")
        error_file_dir_not_found: Final = parse_cli_args(
            [__file__, "-file-dir", str(file_dir_not_found)]
        ).error
        # Don't throw errors if file exists
        self.assertIsNone(parse_cli_args([*self.valid_args]).error)
        # Don't throw errors if additional file directory exists
        self.assertIsNone(
            parse_cli_args(
                [*self.valid_args, "-file-dir", str(Path(__file__).parent)]
            ).error
        )

    def test_parsed_cli_args_same(self):
        for cli_args, result, expected in zip(
            self.cli_args, self.results, self.expected
        ):
            with self.subTest(cli_args=cli_args):
                self.assertEqual(
                    result,
                    expected,
                    f"Check if parsed cli args {result=}=={expected=}",
                )



#!/usr/bin/env python3

# Internal packages
import logging
import re
import textwrap
from pathlib import Path
from re import Match
from typing import Optional, Dict, List, Final

# Installed packages
import markdown
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import get_formatter_by_name
from pygments.util import ClassNotFound

# Local modules
from md2anki.create_id import create_unique_id
from md2anki.html_util import fix_inline_code_p_tags
from md2anki.md_util import (
    md_update_local_filepaths,
    md_update_code_parts,
    md_update_images,
    md_update_math_sections,
    md_update_generic_id_sections,
)
from md2anki.subprocess import (
    subprocess_evaluate_code,
    UnableToEvaluateCodeException,
)

log = logging.getLogger(__name__)

REGEX_MATH_BLOCK: Final = re.compile(r"\$\$([\S\s\n]+?)\$\$", flags=re.MULTILINE)
REGEX_MD_COMMENT_INSERT_FILE: Final = re.compile(
    r"(?:^[ \t]*[\n\r]|\A)(?P<indent>[ \t]*)\[//]:\s+#\s+\(MD2ANKI_INSERT_FILE=(?P<file>.*?)\)(?=(?:[ \t]*[\n\r]){2,}|[ \t]*\Z)",
    flags=re.MULTILINE,
)


def multi_line_math_block_to_single_line(regex_group_match: Match):
    return "".join(regex_group_match.group().splitlines())


def update_md_image_to_html(
    file_path: str,
    file_description: str,
    width: Optional[str],
    height: Optional[str],
) -> str:
    style = ""
    if width is not None:
        style += f"width: {width};"
    if height is not None:
        style += f"height: {height};"
    return f'<img src="{file_path}" alt="{file_description}" style="{style}">'


def md_preprocessor_md2anki(
    md_content: str,
    dir_dynamic_files: Path,
    custom_program: Dict[str, List[str]],
    custom_program_args: Dict[str, List[List[str]]],
    external_file_dirs: List[Path],
    evaluate_code: bool = False,
    keep_temp_files: bool = False,
    anki_latex_math: bool = False,
    render_to_html: bool = False,
) -> str:
    """
    Args:
        @param md_content: Markdown content with md2anki/pandoc/... macros
        @param dir_dynamic_files: Directory for dynamically created files
        @param custom_program: Custom program commands
        @param custom_program_args: Custom program command arguments
        @param evaluate_code: Evaluate code
        @param keep_temp_files: Keep temporary files (debugging)
        @param anki_latex_math: Additionally render certain macros with HTML in mind
        @param render_to_html: Additionally render certain macros with HTML in mind
    Returns:
        The update markdown content.
    """

    log.debug(f">> Update md2anki macro Markdown to native Markdown")
    log.debug(f"   > {md_content=}")

    def insert_external_file(regex_group_match: Match):
        indent = regex_group_match.group(1)
        filepath = Path(regex_group_match.group(2))
        found_file = False
        if filepath.is_file():
            found_file = True
        else:
            for external_file_dir in external_file_dirs:
                if external_file_dir.joinpath(filepath).is_file():
                    found_file = True
                    filepath = external_file_dir.joinpath(filepath)
                    break
        if not found_file:
            raise RuntimeError(
                f"File to be inserted was not found ({filepath=}, {external_file_dirs=})"
            )
        with open(filepath, "r") as f:
            return f"\n\n{textwrap.indent(f.read(), indent)}\n"

    md_content = re.sub(REGEX_MD_COMMENT_INSERT_FILE, insert_external_file, md_content)

    log.debug(f">> Updated insert file comments")
    log.debug(f"   > {md_content=}")

    # Find and store all code sections so that they won't be changed by other changes
    code_sections: Dict[int, str] = dict()
    placeholder_code_section: Final = (
        f"<div>md2anki_placeholder_start_code_{create_unique_id()}",
        f"md2anki_placeholder_end_code_{create_unique_id()}</div>",
    )

    def update_code_section_with_images_or_placeholder(
        code: str,
        code_block: bool,
        language: Optional[str],
        code_block_indent: Optional[str],
    ):
        # Fix pandoc inline notation for language formatting
        if language is not None and language.startswith("."):
            language = language[1:]
        indent = code_block_indent if code_block_indent is not None else ""
        # Fix code block condent if indented
        indent_free_code = code
        if code_block and code_block_indent is not None and len(code_block_indent) > 0:
            indent_free_code = textwrap.dedent(code)
        prefix = "\n" if code_block else ""
        # Detect executable code
        try:
            if evaluate_code and language is not None and language.startswith("="):
                code_output, image_list = subprocess_evaluate_code(
                    language[1:],
                    indent_free_code,
                    dir_dynamic_files=dir_dynamic_files,
                    custom_program=custom_program,
                    custom_program_args=custom_program_args,
                    keep_temp_files=keep_temp_files,
                    customEnv={
                        "MD2ANKI_ANKI_HTML": str(render_to_html),
                        "MD2ANKI_PANDOC_PDF": str(not render_to_html),
                    },
                )
                log.debug(
                    f"> Evaluate {indent_free_code=}: {code_output=}, {image_list=}",
                )
                if len(image_list) > 0:
                    return prefix + "\n".join(
                        map(lambda x: f"{indent}![]({x})\n", image_list)
                    )
                else:
                    return prefix + textwrap.indent("".join(code_output), indent)
            elif language is not None and language.startswith("="):
                # If there is no code update values
                language = language[1:]
                log.warning(
                    f"Code ({language!r}) for evaluation was found but evaluation is disabled!",
                )
        except UnableToEvaluateCodeException as err:
            log.error(err)
        if language is None:
            language = "text"
        if render_to_html is False:
            if code_block:
                return prefix + f"{indent}```{language}\n{code}```\n"
            else:
                return f"`{code}`" + (
                    ("{." + language + "}") if language is not None else ""
                )
        try:
            language_lexer = get_lexer_by_name(language)
        except ClassNotFound as err:
            log.warning(
                f"Default to text lexer ({language=}, {err=})",
            )
            language_lexer = get_lexer_by_name("text")
        html_formatter = get_formatter_by_name("html", noclasses=True)
        pygments_html_output = highlight(code, language_lexer, html_formatter).replace(
            "background: #f8f8f8", ""
        )
        if code_block:
            code_section = textwrap.indent(
                pygments_html_output.replace(
                    'class="highlight"', 'class="highlight highlight_block"'
                ),
                indent,
            )
        else:
            code_section = (
                pygments_html_output.replace("\n", " ")
                .replace("\r", "")
                .replace('class="highlight"', 'class="highlight highlight_inline"')
                .rstrip()
            )

        code_section_index = len(code_sections)
        code_sections[code_section_index] = code_section
        return f"{prefix}{placeholder_code_section[0]}{code_section_index}{placeholder_code_section[1]}"

    md_content = md_update_code_parts(
        md_content, update_code_section_with_images_or_placeholder
    )

    log.debug(f">> Updated code sections with images or placeholders({code_sections=})")
    log.debug(f"   > {md_content=}")

    md_content = re.sub(
        REGEX_MATH_BLOCK, multi_line_math_block_to_single_line, md_content
    )

    # Collect all math sections to later overwrite them again
    math_sections: Dict[int, str] = dict()
    placeholder_math_section: Final = (
        f"md2anki_placeholder_start_math_{create_unique_id()}",
        f"md2anki_placeholder_end_math_{create_unique_id()}",
    )

    if anki_latex_math is True:

        def update_math_section_with_placeholder(math_section: str, block: bool) -> str:
            math_section_index = len(math_sections)
            math_fence = ("\\[", "\\]") if block else (f"\\(", "\\)")
            math_sections[
                math_section_index
            ] = f"{math_fence[0]}{math_section}{math_fence[1]}"
            return f"{placeholder_math_section[0]}{math_section_index}{placeholder_math_section[1]}"

        md_content = md_update_math_sections(
            md_content, update_math_section_with_placeholder
        )

        log.debug(f">> Updated math sections with placeholders ({math_sections=})")
        log.debug(f"   > {md_content=}")

    # Reset local file paths to resources
    md_content = md_update_local_filepaths(md_content)

    log.debug(f">> Reset local filepaths")
    log.debug(f"   > {md_content=}")

    if render_to_html is True:
        # Extract files that this card requests and update paths
        md_content = md_update_images(md_content, update_md_image_to_html)

        log.debug(f">> Updated images to HTML")
        log.debug(f"   > {md_content=}")

        # Render many elements like tables by converting MD to HTML
        md_content = markdown.markdown(
            md_content,
            extensions=["markdown.extensions.extra"],
            tab_length=2,
        )

        log.debug(">> Render markdown to HTML")
        log.debug(f"   > {md_content=}")

    if anki_latex_math is True:
        # Insert all math sections again
        def update_math_section_placeholder_with_value(placeholder_index: str) -> str:
            return math_sections.pop(int(placeholder_index))

        md_content = md_update_generic_id_sections(
            md_content,
            placeholder_math_section,
            update_math_section_placeholder_with_value,
        )

        log.debug(f">> Replaced math section placeholders")
        log.debug(f"   > {md_content=}")

        if len(math_sections) > 0:
            raise RuntimeError(
                f"Not all math sections were inserted back! ({math_sections=})"
            )

    # Insert all code sections again
    def update_code_section_placeholder_with_value(placeholder_index: str) -> str:
        return code_sections.pop(int(placeholder_index))

    md_content = md_update_generic_id_sections(
        md_content,
        placeholder_code_section,
        update_code_section_placeholder_with_value,
    )

    log.debug(f">> Replaced code section placeholders")
    log.debug(f"   > {md_content=}")

    if len(code_sections) > 0:
        raise RuntimeError(
            f"Not all code sections were inserted back! ({code_sections=})"
        )

    if render_to_html is True:
        # Postfix for HTML p tags in front of inline code
        md_content = fix_inline_code_p_tags(md_content)

    return md_content



#!/usr/bin/env python3

# Internal packages
import logging
import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import List, Optional, Final, Dict, Tuple

# Local modules
from md2anki.info import md2anki_name

log = logging.getLogger(__name__)


class UnableToEvaluateCodeException(Exception):
    """Raised when not able to evaluate code"""

    pass


class ProgramNotFoundException(Exception):
    """Raised when a program cannot be found"""

    pass


class ProgramExitedWithWarningsException(Exception):
    """Raised when a program exists with warnings"""

    pass


def run_subprocess(
    command: str,
    arguments: Optional[List[str]] = None,
    cwd: Optional[Path] = None,
    customEnv: Optional[Dict[str, str]] = None,
):
    command_path_global: Final = shutil.which(command)
    command_path_local = (
        command_path_global if cwd is None else shutil.which(str(cwd.joinpath(command)))
    )
    if (
        command_path_global is None
        and command_path_local is None
        and Path(command).is_file()
        and os.access(command, os.X_OK)
    ):
        command_path_local = command
    command_path: Final = (
        command_path_local if command_path_local is not None else command_path_global
    )
    if command_path is None:
        raise ProgramNotFoundException(f"{command=} could not be found")
    if arguments is None:
        arguments = list()
    log.debug(f"Run subprocess {command=}/{command_path=} {arguments=} ({cwd=})")
    my_env = os.environ.copy()
    if customEnv is not None:
        for key, value in customEnv.items():
            my_env[key] = value
    p = subprocess.run(
        [command_path, *arguments], capture_output=True, text=True, cwd=cwd, env=my_env
    )
    log.debug(f"Subprocess output {p.stdout=} {p.returncode=}")
    if p.returncode != 0:
        raise ProgramExitedWithWarningsException(
            f"Program exited with warnings ({command=},{arguments=},{cwd=},"
            f"{p.returncode=},{p.stdout=},{p.stderr=})"
        )
    return p.stdout


DEFAULT_CODE_NAME: Final = "MD2ANKI_CODE"
DEFAULT_CODE_FILE_NAME_START: Final = "MD2ANKI_CODE_FILE="
DEFAULT_CUSTOM_PROGRAM: Final[Dict[str, List[Tuple[str, List[str]]]]] = {
    "py": [("python", ["-c", DEFAULT_CODE_NAME])],
    "js": [("node", ["-e", DEFAULT_CODE_NAME])],
    "ts": [("ts-node", [f"{DEFAULT_CODE_FILE_NAME_START}code.ts"])],
    "pl": [
        (
            "swipl",
            [
                "-O",
                "-s",
                f"{DEFAULT_CODE_FILE_NAME_START}code.pl",
                "-g",
                "true",
                "-t",
                "halt.",
            ],
        )
    ],
    "latex": [
        (
            "latexmk",
            [
                "-shell-escape",
                "-pdf",
                f"{DEFAULT_CODE_FILE_NAME_START}code.tex",
            ],
        ),
        (
            "inkscape",
            ["--export-filename=code.svg", "code.pdf"],
        ),
    ],
    "cpp": [
        (
            "clang++",
            [
                "-Wall",
                "-std=c++20",
                f"{DEFAULT_CODE_FILE_NAME_START}main.cpp",
                "-o",
                "main.exe",
            ],
        ),
        ("main.exe", []),
    ],
    "c": [
        (
            "clang",
            ["-std=c17", f"{DEFAULT_CODE_FILE_NAME_START}main.c", "-o", "main.exe"],
        ),
        ("main.exe", []),
    ],
}


def subprocess_evaluate_code(
    program: str,
    code: str,
    custom_program: Dict[str, List[str]],
    custom_program_args: Dict[str, List[List[str]]],
    dir_dynamic_files: Optional[Path] = None,
    keep_temp_files: bool = False,
    customEnv: Optional[Dict[str, str]] = None,
) -> Tuple[List[str], List[Path]]:
    """Return the command outputs and the found images."""
    log.debug(f"Evaluate code ({program=},{code=})")

    dir_path_temp = Path(tempfile.mkdtemp(prefix=f"{md2anki_name}_evaluate_code_"))

    if program in custom_program:
        program_binaries = custom_program[program]
    else:
        raise UnableToEvaluateCodeException(f"Unsupported {program=} ({code=})")
    if program in custom_program_args:

        def insert_code_or_code_file(program_bin_arg: str) -> str:
            if program_bin_arg == DEFAULT_CODE_NAME:
                return code
            if program_bin_arg.startswith(DEFAULT_CODE_FILE_NAME_START):
                code_file_name = program_bin_arg[len(DEFAULT_CODE_FILE_NAME_START) :]
                if len(code_file_name) == 0:
                    raise RuntimeError(
                        f"Custom code file name had 0 length: {program_bin_arg!r}"
                    )
                with open(dir_path_temp.joinpath(code_file_name), "w") as temp_file:
                    temp_file.write(code)
                return code_file_name
            return program_bin_arg

        def update_list(program_bin_arg_dict_entry: List[str]) -> List[str]:
            return list(map(insert_code_or_code_file, program_bin_arg_dict_entry))

        program_binaries_args: List[List[str]] = list(
            map(update_list, custom_program_args[program])
        )
    else:
        program_binaries_args = list()
    while len(program_binaries_args) < len(program_binaries):
        program_binaries_args.append(list())

    try:
        results: List[str] = list()
        if len(program_binaries) != len(program_binaries_args):
            raise RuntimeError(
                f"Program binary ({len(program_binaries)}) and args list ({len(program_binaries_args)}) had different sizes"
            )
        for program_binary, program_binary_args in zip(
            program_binaries, program_binaries_args
        ):
            results.append(
                run_subprocess(
                    program_binary,
                    program_binary_args,
                    cwd=dir_path_temp,
                    customEnv=customEnv,
                )
            )
        images_list = list()
        if dir_dynamic_files is not None:
            for supported_media_files in [".svg", ".png", ".jpg"]:
                for glob_file in dir_path_temp.rglob(f"*{supported_media_files}"):
                    log.debug(
                        f"Copy created file {glob_file=} to {dir_dynamic_files=}",
                    )
                    shutil.copy2(glob_file, dir_dynamic_files)
                    images_list.append(dir_dynamic_files.joinpath(glob_file.name))
            log.debug(f"{images_list=} {results=}")
        return results, images_list

    finally:
        if not keep_temp_files:
            shutil.rmtree(dir_path_temp)



#!/usr/bin/env python3

# Internal packages
from dataclasses import dataclass
from pathlib import Path
from typing import Final

RELATIVE_RES_DIR: Final = Path("res")
RELATIVE_CSS_FILE_PATH: Final = RELATIVE_RES_DIR.joinpath("stylesheet.css")
RELATIVE_CSS_FILE_PATH_TYPE_ANSWER: Final = RELATIVE_RES_DIR.joinpath(
    "stylesheet_type_answer.css"
)


@dataclass
class ProgramVersion:
    """
    Contains all information about the md2anki version.
    """

    major: int
    minor: int = 0
    patch: int = 0
    beta: bool = False

    def __repr__(self):
        return f"{self.major}.{self.minor}.{self.patch}{'b' if self.beta else ''}"


md2anki_name: Final = "md2anki"
md2anki_author: Final = "AnonymerNiklasistanonym"
md2anki_description: Final = "Convert Markdown formatted documents to anki decks"
md2anki_url: Final = "https://github.com/AnonymerNiklasistanonym/Md2Anki"
md2anki_url_git: Final = f"{md2anki_url}.git"
md2anki_url_bug_tracker: Final = f"{md2anki_url}/issues"
md2anki_version: Final = ProgramVersion(major=3, minor=0, patch=12, beta=True)

MD_ANKI_DECK_HEADING_SUBDECK_PREFIX: Final = "Subdeck: "
MD_ANKI_NOTE_QUESTION_ANSWER_SEPARATOR: Final = "---"
ANKI_SUBDECK_SEPARATOR: Final = "::"



#!/usr/bin/env python3

# Future import to support Python 3.9
from __future__ import annotations

# Internal packages
import argparse
import json
import logging
import sys
from dataclasses import dataclass, field
from operator import attrgetter
from pathlib import Path
from typing import Optional, List, Dict, Final, Tuple, TypeVar, Callable

# Local modules
from md2anki.info import md2anki_version, md2anki_name
from md2anki.md_to_pdf import PANDOC_ARGS_PDF
from md2anki.note_models import AnkiCardModelId
from md2anki.subprocess import DEFAULT_CUSTOM_PROGRAM


class SortedArgumentDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def add_arguments(self, actions):
        actions = sorted(actions, key=attrgetter("option_strings"))
        super(argparse.ArgumentDefaultsHelpFormatter, self).add_arguments(actions)


class MdInputFileNotFoundException(Exception):
    """Raised when a markdown input file is not found"""

    md_input_file_path: Final[Path]

    def __init__(self, md_input_file_path: Path):
        absolute_path: Final = md_input_file_path.absolute()
        super().__init__(
            f"Markdown input file was not found ({md_input_file_path!r}, {absolute_path=})"
        )
        self.md_input_file_path = md_input_file_path


@dataclass
class Md2AnkiArgs:
    """
    Contains all information of the md2anki command line arguments.
    """

    additional_file_dirs: List[Path] = field(default_factory=lambda: list())
    """Additional file directories for external file references (i.e. images)."""
    anki_card_model: AnkiCardModelId = AnkiCardModelId.DEFAULT
    """Anki card model."""
    md_input_file_paths: List[Path] = field(default_factory=lambda: list())
    """The markdown input file paths."""
    md_output_file_paths: Optional[List[Path]] = None
    """The optional output file paths of the updated markdown files (i.e. with added IDs)."""
    md_output_dir_path: Optional[Path] = None
    """The optional output dir path of the updated markdown files (i.e. with added IDs)."""
    anki_output_file_path: Optional[Path] = None
    """The output file path of the anki deck output file."""
    backup_output_dir_path: Optional[Path] = None
    """The output dir path of the backup of the anki deck."""
    pdf_output_file_path: Optional[Path] = None
    """The output file path of the anki deck."""
    md_heading_depth: int = 1
    """The default heading depth for the anki deck heading (increases for each subdeck/question by 1)."""
    custom_program: Dict[str, List[str]] = field(default_factory=lambda: dict())
    """Custom programs for languages used for code evaluation."""
    custom_program_args: Dict[str, List[List[str]]] = field(
        default_factory=lambda: dict()
    )
    """Custom program args for languages used for code evaluation."""

    evaluate_code = False
    """Evaluate code."""
    keep_temp_files = False
    """Remove temporary files."""
    error: Optional[MdInputFileNotFoundException] = None
    """Error message in case there was an error parsing CLI args."""


T = TypeVar("T")
U = TypeVar("U")


def convert_list_to_dict_merged(
    list_object: List[Tuple[str, T]],
    value_func: Callable[[T], U],
) -> Dict[str, List[U]]:
    dict_object: Dict[str, List[U]] = dict()
    for key, value in list_object:
        dict_object.setdefault(key, []).append(value_func(value))
    return dict_object


DEFAULT_CUSTOM_PROGRAMS: Final = [
    (key, program)
    for key, values in DEFAULT_CUSTOM_PROGRAM.items()
    for program, _ in values
] + [(key, program) for key, values in PANDOC_ARGS_PDF.items() for program, _ in values]
DEFAULT_CUSTOM_PROGRAM_ARGS: Final = [
    (key, json.dumps(program_args))
    for key, values in DEFAULT_CUSTOM_PROGRAM.items()
    for _, program_args in values
] + [
    (key, json.dumps(program_args))
    for key, values in PANDOC_ARGS_PDF.items()
    for _, program_args in values
]

log = logging.getLogger(__name__)


def str_to_str(a: str) -> str:
    return a


def json_str_to_str_list(a: str) -> List[str]:
    possible_list: List[str] = json.loads(a)
    if not all(isinstance(elem, str) for elem in possible_list):
        raise RuntimeError(
            f"Custom program args need to be a JSON string list ({possible_list!r})"
        )
    return possible_list


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Create an anki deck file (.apkg) from one or more Markdown documents. "
        "If no custom output path is given the file name of the document (+ .apkg) is used.",
        formatter_class=SortedArgumentDefaultsHelpFormatter,
        prog=md2anki_name,
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"%(prog)s {md2anki_version}"
    )

    parser.add_argument(
        "-d",
        "--debug",
        dest="log_level",
        const="DEBUG",
        default="INFO",
        action="store",
        nargs="?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="custom log level to the console",
    )

    parser.add_argument(
        "-e",
        "--evaluate-code",
        action="store_true",
        help="evaluate code parts that start with an '=' like '`print(1+1)`{=python}",
    )
    parser.add_argument(
        "-k",
        "--keep-temp-files",
        action="store_true",
        help="keep temporary files",
    )
    parser.add_argument(
        "-anki-model",
        metavar="MODEL_ID",
        type=AnkiCardModelId,
        choices=list(AnkiCardModelId),
        default=AnkiCardModelId.DEFAULT,
        help=f"custom anki card model (%(choices)s)",
    )
    parser.add_argument(
        "-o-anki",
        metavar="APKG_FILE",
        type=Path,
        help="custom anki deck (.apkg) output file path [if not given: md input file "
        "name + .apkg]",
    )
    parser.add_argument(
        "-o-md",
        metavar="MD_FILE",
        type=Path,
        help="custom updated (and merged if multiple input files) Markdown (.md) output file path for all input files",
    )
    parser.add_argument(
        "-o-md-dir",
        metavar="MD_DIR",
        type=Path,
        help="custom output directory for all updated Markdown (.md) input files",
    )
    parser.add_argument(
        "-o-backup-dir",
        metavar="BACKUP_DIR",
        type=Path,
        help="create a backup of the anki deck (i.e. merges input files and copies external files) in a directory",
    )
    parser.add_argument(
        "-o-pdf",
        metavar="PDF_FILE",
        type=Path,
        help="create a PDF (.pdf) file of the anki deck (i.e. merges input files and removes IDs)",
    )
    parser.add_argument(
        "-log-file",
        metavar="LOG_FILE",
        type=Path,
        help="log all messages to a text file (.log)",
    )
    parser.add_argument(
        "-file-dir",
        metavar="FILE_DIR",
        type=Path,
        action="extend",
        nargs="*",
        default=[],
        help="add directories that should be checked for referenced files (like relative path images)",
    )

    def check_heading_depth(x: str) -> int:
        x_num = int(x)
        if x_num < 1:
            raise argparse.ArgumentTypeError("Minimum heading depth is 1")
        return x_num

    parser.add_argument(
        "-md-heading-depth",
        metavar="HEADING_DEPTH",
        type=check_heading_depth,
        default=1,
        help="use a custom Markdown heading depth (>=1)",
    )
    parser.add_argument(
        "-custom-program",
        metavar=("language", "program"),
        type=str,
        action="append",
        nargs=2,
        default=DEFAULT_CUSTOM_PROGRAMS,
        help="use custom program for code evaluation",
    )
    parser.add_argument(
        "-custom-program-args",
        metavar=("language", "program-args"),
        type=str,
        action="append",
        nargs=2,
        default=DEFAULT_CUSTOM_PROGRAM_ARGS,
        help="use custom program args for code evaluation",
    )
    parser.add_argument(
        "md_input_files",
        metavar="MD_INPUT_FILE",
        type=Path,
        nargs="+",
        help="Markdown (.md) input file that contains anki deck notes",
    )
    return parser


class FormatterCleanInfo(logging.Formatter):
    def format(self, record):
        if record.levelno == logging.INFO:
            self._style._fmt = "%(message)s"
        else:
            self._style._fmt = "%(asctime)s %(levelname)s %(name)s: %(message)s"
        return super().format(record)


class Formatter(logging.Formatter):
    def format(self, record):
        self._style._fmt = "%(asctime)s %(levelname)s %(name)s: %(message)s"
        return super().format(record)


def parse_cli_args(cli_args: List[str]) -> Md2AnkiArgs:
    """
    Parse the supplied CLI arguments.

    @return: Object that contains all parsed information.
    """
    parser = get_argument_parser()
    args = parser.parse_args(cli_args)
    parsed_args = Md2AnkiArgs()

    handlers: List[logging.FileHandler | logging.StreamHandler] = []

    # Create a logging handler to stdout only for info
    hdlr_info = logging.StreamHandler(sys.stdout)
    hdlr_info.setFormatter(FormatterCleanInfo())
    hdlr_info.setLevel(getattr(logging, args.log_level))
    hdlr_info.addFilter(lambda record: record.levelno < logging.WARNING)
    handlers.append(hdlr_info)

    # Create a logging handler to stderr for warnings and errors
    hdlr_warn_err = logging.StreamHandler(sys.stderr)
    hdlr_warn_err.setFormatter(Formatter())
    hdlr_warn_err.setLevel(logging.WARNING)
    hdlr_warn_err.addFilter(lambda record: record.levelno >= logging.WARNING)
    handlers.append(hdlr_warn_err)

    if args.log_file is not None:
        # Create a logging handler to a file for everything
        hdlr_file = logging.FileHandler(args.log_file, mode="w")
        hdlr_file.setFormatter(Formatter())
        hdlr_file.setLevel(logging.DEBUG)
        handlers.append(hdlr_file)

    # This is necessary to fix the file handler
    logging.root.handlers = []

    # Setup global logging
    logging.basicConfig(
        level=logging.DEBUG,
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=handlers,
    )

    log.debug(f"{cli_args=}, {args=}")

    parsed_args.additional_file_dirs = args.file_dir
    parsed_args.anki_card_model = args.anki_model
    parsed_args.md_heading_depth = args.md_heading_depth
    parsed_args.md_input_file_paths = args.md_input_files
    parsed_args.backup_output_dir_path = args.o_backup_dir
    parsed_args.pdf_output_file_path = args.o_pdf
    parsed_args.evaluate_code = args.evaluate_code

    if args.keep_temp_files is not None:
        parsed_args.keep_temp_files = args.keep_temp_files

    temp_custom_program = convert_list_to_dict_merged(
        args.custom_program[: len(DEFAULT_CUSTOM_PROGRAMS)], str_to_str
    )
    if args.custom_program is not None:
        parsed_args.custom_program = convert_list_to_dict_merged(
            args.custom_program[len(DEFAULT_CUSTOM_PROGRAMS) :], str_to_str
        )
    parsed_args.custom_program = {
        **temp_custom_program,
        **parsed_args.custom_program,
    }

    temp_custom_program_args = convert_list_to_dict_merged(
        args.custom_program_args[: len(DEFAULT_CUSTOM_PROGRAM_ARGS)],
        json_str_to_str_list,
    )
    if args.custom_program_args is not None:
        parsed_args.custom_program_args = convert_list_to_dict_merged(
            args.custom_program_args[len(DEFAULT_CUSTOM_PROGRAM_ARGS) :],
            json_str_to_str_list,
        )
    parsed_args.custom_program_args = {
        **temp_custom_program_args,
        **parsed_args.custom_program_args,
    }

    # If an input file is not found throw error
    for md_input_file_path in parsed_args.md_input_file_paths:
        if not md_input_file_path.is_file():
            parsed_args.error = MdInputFileNotFoundException(md_input_file_path)
            return parsed_args

    # If an additional file directory is not found throw error
    for additional_file_dir in parsed_args.additional_file_dirs:
        if not additional_file_dir.is_dir():
            log.warning(
                f"Ignore {additional_file_dir=!r} because it's not a directory!"
            )
            parsed_args.additional_file_dirs.remove(additional_file_dir)

    parsed_args.anki_output_file_path = (
        args.o_anki
        if args.o_anki
        else parsed_args.md_input_file_paths[0].parent.joinpath(
            f"{parsed_args.md_input_file_paths[0].stem}.apkg"
        )
    )

    if args.o_md:
        parsed_args.md_output_file_paths = [args.o_md]
    elif args.o_md_dir:
        parsed_args.md_output_dir_path = args.o_md_dir
    else:
        parsed_args.md_output_file_paths = parsed_args.md_input_file_paths.copy()

    return parsed_args



#!/usr/bin/env python3

# Internal packages
import logging
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import Set, Optional, List, Tuple, Dict, Final

# Installed packages
import genanki

# Local modules
from md2anki.anki_model import AnkiModel
from md2anki.anki_note import AnkiNote, MdSection
from md2anki.create_id import create_unique_id_int
from md2anki.info import (
    ANKI_SUBDECK_SEPARATOR,
    md2anki_url_git,
    md2anki_name,
    md2anki_version,
    MD_ANKI_DECK_HEADING_SUBDECK_PREFIX,
)
from md2anki.md_util import md_get_used_md2anki_tags

ANKI_SUPPORTED_FILE_FORMATS: Final = [".jpg", ".png", ".svg"]

log = logging.getLogger(__name__)


@dataclass
class AnkiDeck:
    """
    Contains all information of an anki deck
    """

    name: str = "No name"
    """Name of the deck"""
    model: AnkiModel = field(default_factory=lambda: AnkiModel())
    """Model of cards of anki deck"""
    guid: int = create_unique_id_int()
    """Unique id of anki deck"""
    description: str = ""
    """Description of anki deck"""
    notes: List[AnkiNote] = field(default_factory=lambda: list())
    """List of anki notes"""
    additional_file_dirs: List[Path] = field(default_factory=lambda: list())
    """List of additional file dirs that should be searched for missing files used in notes"""
    tags: Set[str] = field(default_factory=lambda: set())
    """List of anki tags"""
    md_document_order: int = -1
    """The order of the anki deck in the read md document"""

    def genanki_create_deck(
        self,
        anki_card_model: genanki.Model,
        dir_dynamic_files: Path,
        custom_program: Dict[str, List[str]],
        custom_program_args: Dict[str, List[List[str]]],
        external_file_dirs: List[Path],
        evaluate_code: bool,
        keep_temp_files: bool,
    ) -> genanki.Deck:
        tmp_anki_deck = genanki.Deck(self.guid, self.name)
        for note in self.notes:
            tmp_anki_deck.add_note(
                note.genanki_create_note(
                    anki_card_model=anki_card_model,
                    dir_dynamic_files=dir_dynamic_files,
                    custom_program=custom_program,
                    custom_program_args=custom_program_args,
                    evaluate_code=evaluate_code,
                    external_file_dirs=external_file_dirs,
                    keep_temp_files=keep_temp_files,
                )
            )
        return tmp_anki_deck

    def get_used_global_tags(self) -> Set[str]:
        """Get the used global tags of the anki deck"""
        return self.tags.union(md_get_used_md2anki_tags(self.description))

    def get_local_files_from_notes(self) -> List[Path]:
        files: Set[Path] = set()
        for note in self.notes:
            files.update(note.get_used_local_files())
        file_list: List[Path] = list()
        for file in files:
            # Check if every file can be found
            if file.is_file():
                file_list.append(file)
            else:
                file_found = False
                # Check if file is located in one of the additional file dirs
                for additional_file_dir in self.additional_file_dirs:
                    file_path = additional_file_dir.joinpath(file)
                    log.debug(
                        f"{file=} not found check in {additional_file_dir=}: {file_path=}"
                    )
                    if file_path.is_file():
                        file_list.append(file_path)
                        file_found = True
                        break
                if not file_found:
                    raise Exception(f"{file=} not found ({self.additional_file_dirs=})")
        return file_list

    def genanki_create_anki_deck(
        self,
        dir_dynamic_files: Path,
        custom_program: Dict[str, List[str]],
        custom_program_args: Dict[str, List[List[str]]],
        external_file_dirs: List[Path],
        evaluate_code: bool = False,
        keep_temp_files: bool = False,
    ) -> Tuple[genanki.Deck, List[Path]]:
        """Return anki deck and a list of all media files."""
        genanki_anki_deck = self.genanki_create_deck(
            self.model.genanki_create_model(),
            dir_dynamic_files=dir_dynamic_files,
            custom_program=custom_program,
            custom_program_args=custom_program_args,
            evaluate_code=evaluate_code,
            external_file_dirs=external_file_dirs,
            keep_temp_files=keep_temp_files,
        )
        media_files = self.get_local_files_from_notes()
        # Add all supported dynamic media files
        for supported_file_format in ANKI_SUPPORTED_FILE_FORMATS:
            media_files.extend(dir_dynamic_files.rglob(f"*{supported_file_format}"))
        return genanki_anki_deck, media_files

    def create_md_sections(
        self, local_asset_dir_path: Optional[Path] = None
    ) -> Tuple[MdSection, List[MdSection]]:
        output: List[MdSection] = []
        for note in self.notes:
            output.append(
                note.create_md_section(local_asset_dir_path=local_asset_dir_path)
            )
        return (
            MdSection(self.name, str(self.guid), self.description),
            output,
        )


def genanki_package_anki_decks_to_file(
    anki_decks: List[Tuple[genanki.Deck, List[Path]]], output_file_path: Path
):
    anki_deck_package = genanki.Package([a[0] for a in anki_decks])
    media_files = [a[1] for a in anki_decks]
    anki_deck_package.media_files = [
        item for sublist in media_files for item in sublist
    ]
    anki_deck_package.write_to_file(output_file_path)
    log.debug(
        f"Write anki decks {', '.join([a[0].name + ' (' + str(len(a[0].notes)) + ')' for a in anki_decks])} to "
        f"'{output_file_path}' [.apkg]"
    )


def md_merge_anki_decks_to_md_file(
    anki_decks: List[AnkiDeck],
    output_file_path: Path,
    local_asset_dir_path: Optional[Path] = None,
    initial_heading_depth: int = 1,
    remove_ids=False,
):
    """Merge anki decks to a single markdown output file"""
    heading_stack: List[Tuple[str, str]] = []
    """Store what headings were currently written"""
    first_heading = True
    with open(output_file_path, "w", encoding="utf-8") as file:
        for anki_deck in anki_decks:
            # Create a markdown section for the notes and the heading
            anki_deck_heading, anki_deck_notes = anki_deck.create_md_sections(
                local_asset_dir_path=local_asset_dir_path
            )
            # Determine the heading depth
            heading_names = anki_deck_heading.heading.split(ANKI_SUBDECK_SEPARATOR)
            heading_depth = len(heading_names)
            # Check if heading was already written
            heading_same_as_previous_heading = [
                a[0] for a in heading_stack
            ] == heading_names
            heading_parent_deck_of_previous_heading = (
                len(heading_names) < len(heading_stack)
                and [a[0] for a in heading_stack[: len(heading_names)]] == heading_names
            )
            heading_description_same = (
                len(heading_stack) >= len(heading_names)
                and heading_stack[len(heading_names) - 1][1]
                == anki_deck_heading.content
            )
            if not (
                (heading_same_as_previous_heading and heading_description_same)
                or (
                    heading_parent_deck_of_previous_heading and heading_description_same
                )
            ):
                # Fix heading name
                if heading_depth > 1:
                    anki_deck_heading.heading = heading_names[-1]
                # Write notes and heading to file
                if not first_heading:
                    file.write("\n")
                first_heading = False
                file.write(
                    anki_deck_heading.create_string(
                        depth=heading_depth + initial_heading_depth - 1,
                        heading_prefix=MD_ANKI_DECK_HEADING_SUBDECK_PREFIX
                        if heading_depth > 1
                        else "",
                        with_heading_id=not remove_ids,
                    )
                )
                new_heading_stack = []
                for index, heading_name in enumerate(heading_names):
                    heading_content = ""
                    if index == len(heading_names) - 1:
                        heading_content = anki_deck_heading.content
                    elif (
                        len(heading_stack) >= index - 1
                        and heading_name == heading_stack[index][0]
                    ):
                        heading_content = heading_stack[index][1]
                    new_heading_stack.append((heading_name, heading_content))
                heading_stack = new_heading_stack
            for index, anki_deck_note in enumerate(anki_deck_notes):
                file.write("\n")
                file.write(
                    anki_deck_note.create_string(
                        depth=heading_depth + initial_heading_depth,
                        with_heading_id=not remove_ids,
                    )
                )
    log.debug(
        f"Write anki decks {', '.join([a.name + ' (' + str(len(a.notes)) + ')' for a in anki_decks])} to "
        f"'{output_file_path}' [.md]"
    )


def backup_anki_decks_to_dir(
    anki_decks_list: List[List[AnkiDeck]],
    output_dir_path: Path,
    initial_heading_depth: int = 1,
):
    if not output_dir_path.exists():
        output_dir_path.mkdir()
    asset_dir_path_relative = Path("assets")
    asset_dir_path = output_dir_path.joinpath(asset_dir_path_relative)
    local_assets: List[Path] = []

    # Backup each anki deck
    for index, anki_decks in enumerate(anki_decks_list):
        document_name = (
            f"document_part_{index + 1:02}.md"
            if len(anki_decks_list) > 1
            else "document.md"
        )
        md_merge_anki_decks_to_md_file(
            anki_decks,
            output_file_path=output_dir_path.joinpath(document_name),
            initial_heading_depth=initial_heading_depth,
            local_asset_dir_path=asset_dir_path_relative,
        )
        for anki_deck in anki_decks:
            local_assets.extend(anki_deck.get_local_files_from_notes())

    if len(local_assets) > 0:
        if not asset_dir_path.exists():
            asset_dir_path.mkdir()
        for local_asset in local_assets:
            shutil.copyfile(local_asset, asset_dir_path.joinpath(local_asset.name))
    bash_build: Final = output_dir_path.joinpath("build.sh")
    pwsh_build: Final = output_dir_path.joinpath("build.ps1")
    with open(bash_build, "w", encoding="utf-8") as bash_file, open(
        pwsh_build, "w", encoding="utf-8"
    ) as pwsh_file:
        document_name = "document"
        anki_deck_name = "anki_deck"
        info_string = f"Automatically created using {md2anki_name} {md2anki_version}"
        # Shebang and version
        bash_file.write(f"#!/usr/bin/env bash\n\n# {info_string}\n\n")
        pwsh_file.write(f"#!/usr/bin/env pwsh\n\n# {info_string}\n\n")
        # Git clone
        bash_file.write(
            f'Md2AnkiGitDir="{md2anki_name}"\n'
            'if [ ! -d "$Md2AnkiGitDir" ] ; then\n'
            f'    git clone "{md2anki_url_git}" {md2anki_name}\n'
            "fi\n\n"
        )
        pwsh_file.write(
            f'$Md2AnkiGitDir = Join-Path $PSScriptRoot -ChildPath "{md2anki_name}"\n'
            "if (-not (Test-Path -LiteralPath $Md2AnkiGitDir)) {\n"
            f'    git clone "{md2anki_url_git} {md2anki_name}"\n'
            "}\n\n"
        )
        # Run command
        bash_file.write(f"./$Md2AnkiGitDir/run.sh")
        pwsh_file.write(
            "$Md2AnkiRun = Join-Path $Md2AnkiGitDir -ChildPath run.ps1\n"
            'Invoke-Expression "$Md2AnkiRun'
        )
        # > Document list
        if len(anki_decks_list) > 1:
            for index in range(1, len(anki_decks_list) + 1):
                bash_file.write(f' "{document_name}_part_{index:02}.md"')
                pwsh_file.write(f' `"{document_name}_part_{index:02}.md`"')
        else:
            bash_file.write(f' "{document_name}.md"')
            pwsh_file.write(f' `"{document_name}.md`"')
        # > File directories
        bash_file.write(' -file-dir "."')
        pwsh_file.write(' -file-dir `".`"')
        # > Anki deck output
        bash_file.write(f' -o-anki "{anki_deck_name}.apkg" "$@"\n')
        pwsh_file.write(f' -o-anki `"{anki_deck_name}.apkg`" $args"\n')

    log.debug(f"Write anki deck backup to '{output_dir_path}'")



#!/usr/bin/env python3

# Internal packages
import sys
import unittest
from pathlib import Path
from typing import List, Tuple, Set

# Append the module path for md2anki
sys.path.append(str(Path(__file__).parent.parent.joinpath("src")))

# Local modules
from md2anki.anki_note import AnkiNote, MdSection
from md2anki.cli import (
    DEFAULT_CUSTOM_PROGRAMS,
    convert_list_to_dict_merged,
    str_to_str,
    json_str_to_str_list,
    DEFAULT_CUSTOM_PROGRAM_ARGS,
)
from md2anki.note_models import create_default_anki_deck_model


class TestAnkiNoteGeneric(unittest.TestCase):
    def setUp(self):
        self.anki_note_list: List[AnkiNote] = list()
        self.expected_tags: List[Set[str]] = list()
        self.expected_local_files: List[Set[Path]] = list()
        self.expected_md_sections: List[Tuple[MdSection, MdSection]] = list()
        self.expected_anki_output: List[Tuple[str, str]] = list()
        self.results_tags: List[Set[str]] = list()
        self.results_local_files: List[Set[Path]] = list()
        self.results_md_sections: List[Tuple[MdSection, MdSection]] = list()
        self.results_anki_output: List[Tuple[str, str]] = list()

        anki_card_model = create_default_anki_deck_model()
        dir_dynamic_files = Path(__file__).parent.joinpath("dynamic_files")
        local_asset_dir = Path("assets")

        test_data: List[
            Tuple[
                AnkiNote,
                Set[str],
                Set[Path],
                Tuple[MdSection, MdSection],
                Tuple[str, str],
            ]
        ] = [
            (
                AnkiNote(
                    question="question", answer="answer", tags={"tag"}, guid="1234"
                ),
                {"tag"},
                set(),
                (
                    MdSection(heading="question", heading_id="1234", content="answer"),
                    MdSection(heading="question", heading_id="1234", content="answer"),
                ),
                ("<p>question</p>", "<p>answer</p>"),
            ),
            (
                AnkiNote(
                    question="question\n`{=:note_tag:=}`",
                    answer="answer $66 * 42 = 10$ ![](res/wikipedia_svg_demo.svg)",
                    guid="1234",
                ),
                {"note_tag"},
                {Path("res/wikipedia_svg_demo.svg")},
                (
                    MdSection(
                        heading="question",
                        heading_id="1234",
                        content="`{=:note_tag:=}`\n\n---\n\nanswer $66 * 42 = 10$ ![](res/wikipedia_svg_demo.svg)",
                    ),
                    MdSection(
                        heading="question",
                        heading_id="1234",
                        content="`{=:note_tag:=}`"
                        + f"\n\n---\n\nanswer $66 * 42 = 10$ ![]({local_asset_dir.joinpath('wikipedia_svg_demo.svg')})",
                    ),
                ),
                (
                    '<p class="highlight_inline_next">question</p>\n'
                    + '<div class="highlight highlight_inline" style=""><pre style="line-height: '
                    + '125%;"><span></span>{=:note_tag:=} </pre></div>',
                    '<p>answer \\(66 * 42 = 10\\) <img alt="" src="wikipedia_svg_demo.svg" style=""/></p>',
                ),
            ),
        ]

        for (
            test_input,
            test_expected_tags,
            test_expected_local_files,
            test_expected_md_section,
            test_expected_anki_output,
        ) in test_data:
            self.anki_note_list.append(test_input)
            self.expected_tags.append(test_expected_tags)
            self.expected_local_files.append(test_expected_local_files)
            self.expected_md_sections.append(test_expected_md_section)
            self.expected_anki_output.append(test_expected_anki_output)
            self.results_tags.append(test_input.get_used_md2anki_tags())
            self.results_local_files.append(test_input.get_used_local_files())
            self.results_md_sections.append(
                (
                    test_input.create_md_section(),
                    test_input.create_md_section(local_asset_dir_path=local_asset_dir),
                )
            )
            self.results_anki_output.append(
                tuple(
                    # type: ignore
                    test_input.genanki_create_note(
                        anki_card_model=anki_card_model,
                        dir_dynamic_files=dir_dynamic_files,
                        custom_program=convert_list_to_dict_merged(
                            DEFAULT_CUSTOM_PROGRAMS, str_to_str
                        ),
                        custom_program_args=convert_list_to_dict_merged(
                            DEFAULT_CUSTOM_PROGRAM_ARGS, json_str_to_str_list
                        ),
                        external_file_dirs=[],
                    ).fields
                )
            )

    def test_anki_note_tags_same(self):
        for anki_note, expected_tags, result_tags in zip(
            self.anki_note_list,
            self.expected_tags,
            self.results_tags,
        ):
            with self.subTest(anki_note=anki_note):
                self.assertEqual(
                    expected_tags,
                    result_tags,
                    f"Check if anki note {expected_tags=}=={result_tags=}",
                )

    def test_anki_note_local_files_same(self):
        for anki_note, expected_local_files, result_local_files in zip(
            self.anki_note_list,
            self.expected_local_files,
            self.results_local_files,
        ):
            with self.subTest(anki_note=anki_note):
                self.assertEqual(
                    expected_local_files,
                    result_local_files,
                    f"Check if anki note {expected_local_files=}=={result_local_files=}",
                )

    def test_anki_note_md_section_same(self):
        for anki_note, expected_md_section, result_md_section in zip(
            self.anki_note_list,
            self.expected_md_sections,
            self.results_md_sections,
        ):
            with self.subTest(anki_note=anki_note):
                self.assertEqual(
                    expected_md_section,
                    result_md_section,
                    f"Check if anki note {expected_md_section=}=={result_md_section=}",
                )

    def test_anki_note_anki_output_same(self):
        for anki_note, expected_anki_output, result_anki_output in zip(
            self.anki_note_list,
            self.expected_anki_output,
            self.results_anki_output,
        ):
            with self.subTest(anki_note=anki_note):
                self.assertEqual(
                    expected_anki_output,
                    result_anki_output,
                    f"Check if anki note {expected_anki_output=}=={result_anki_output=}",
                )



import sys
from pathlib import Path
from setuptools import setup, find_packages

# Append the module path for md2anki
sys.path.append(str(Path(__file__).parent.joinpath("src", "md2anki")))

from info import (
    md2anki_version,
    md2anki_name,
    md2anki_author,
    md2anki_description,
    md2anki_url,
    md2anki_url_bug_tracker,
    RELATIVE_CSS_FILE_PATH,
    RELATIVE_CSS_FILE_PATH_TYPE_ANSWER,
)

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name=md2anki_name,
    version=str(md2anki_version),
    author=md2anki_author,
    description=md2anki_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=md2anki_url,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=[
        "beautifulsoup4>=4.11.2",
        "genanki>=0.13.0",
        "Markdown>=3.4.1",
        "Pygments>=2.14.0",
    ],
    package_data={
        md2anki_name: [
            str(RELATIVE_CSS_FILE_PATH),
            str(RELATIVE_CSS_FILE_PATH_TYPE_ANSWER),
        ],
    },
    entry_points={
        "console_scripts": [
            f"{md2anki_name}={md2anki_name}:_main",
        ],
    },
    project_urls={
        "Bug Tracker": md2anki_url_bug_tracker,
    },
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python3

# Internal packages
import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Set, Optional, Dict, List
from urllib.parse import ParseResult

# Installed packages
import genanki

# Local modules
from md2anki.create_id import create_unique_id
from md2anki.preprocessor import md_preprocessor_md2anki
from md2anki.md_util import (
    md_get_used_files,
    md_get_used_md2anki_tags,
    md_update_local_filepaths,
)

log = logging.getLogger(__name__)


@dataclass
class MdSection:
    """
    Represents a markdown section of anki deck notes.
    """

    heading: str = ""
    heading_id: str = ""
    content: str = ""

    def create_string(
        self,
        depth: int = 1,
        heading_prefix: Optional[str] = None,
        with_heading_id=True,
    ) -> str:
        output = (
            f"{'#' * depth} {heading_prefix if heading_prefix is not None else ''}{self.heading}"
            f"{f' ({self.heading_id})' if with_heading_id and len(self.heading_id) > 0 else ''}\n"
        )
        if self.content != "":
            output += f"\n{self.content}\n"
        return output


@dataclass
class AnkiNote:
    """
    Contains all information of an anki deck note.
    """

    question: str = ""
    """Question multi line string"""
    answer: str = ""
    """Answer multi line string"""
    tags: Set[str] = field(default_factory=lambda: set())
    """Tags"""
    guid: str = create_unique_id()
    """Unique id"""

    def get_used_local_files(self) -> Set[Path]:
        return set(
            [
                file
                for file in md_get_used_files(self.question).union(
                    md_get_used_files(self.answer)
                )
                if not isinstance(file, ParseResult)
            ]
        )

    def get_used_md2anki_tags(self) -> Set[str]:
        return (
            md_get_used_md2anki_tags(self.question)
            .union(md_get_used_md2anki_tags(self.answer))
            .union(self.tags)
        )

    def genanki_create_note(
        self,
        anki_card_model: genanki.Model,
        dir_dynamic_files: Path,
        custom_program: Dict[str, List[str]],
        custom_program_args: Dict[str, List[List[str]]],
        external_file_dirs: List[Path],
        evaluate_code: bool = False,
        keep_temp_files: bool = False,
    ) -> genanki.Note:
        """
        Args:
            @param anki_card_model: The card model
            @param dir_dynamic_files: Directory for dynamically created files
            @param custom_program: Custom program commands
            @param custom_program_args: Custom program command arguments
            @param evaluate_code: Evaluate code
            @param external_file_dirs: External file directories to insert files from
            @param keep_temp_files: Keep temporary files (debugging)
        Returns:
            An anki note for genanki
        """
        return genanki.Note(
            guid=self.guid,
            model=anki_card_model,
            fields=[
                md_preprocessor_md2anki(
                    self.question,
                    dir_dynamic_files=dir_dynamic_files,
                    custom_program=custom_program,
                    custom_program_args=custom_program_args,
                    evaluate_code=evaluate_code,
                    external_file_dirs=external_file_dirs,
                    keep_temp_files=keep_temp_files,
                    anki_latex_math=True,
                    render_to_html=True,
                ),
                md_preprocessor_md2anki(
                    self.answer,
                    dir_dynamic_files=dir_dynamic_files,
                    custom_program=custom_program,
                    custom_program_args=custom_program_args,
                    evaluate_code=evaluate_code,
                    external_file_dirs=external_file_dirs,
                    keep_temp_files=keep_temp_files,
                    anki_latex_math=True,
                    render_to_html=True,
                ),
            ],
            tags=list(self.get_used_md2anki_tags().union(self.tags)),
        )

    def create_md_section(
        self, local_asset_dir_path: Optional[Path] = None
    ) -> MdSection:
        newline_count_question = len(self.question.splitlines())
        question_header: str = self.question.splitlines()[0].rstrip()
        question_body: str
        if newline_count_question <= 1:
            question_body = ""
        else:
            question_body = (
                f"\n{''.join(self.question.splitlines(keepends=True)[1:])}\n\n---"
            )
        if local_asset_dir_path is not None:
            question_header = md_update_local_filepaths(
                question_header, local_asset_dir_path
            )
            question_body = md_update_local_filepaths(
                question_body, local_asset_dir_path
            )
            answer = md_update_local_filepaths(self.answer, local_asset_dir_path)
        else:
            answer = self.answer

        return MdSection(
            question_header, self.guid, f"{question_body}\n\n{answer}".strip()
        )
